// GpibServer.c

/*************************************************************************
* Copyright (c) 2003 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. mpfGpib is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
// Author: Marty Kraimer, Thomas Coleman
// Date: 7/98


#include <vxWorks.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <taskLib.h>
#include <semLib.h>
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "gpibServer.h"
#include "Gpib.h"

class GpibServer
{
public:
    GpibServer(const char *name, const char *gpibName, int bufSize, int queueSize);
    static void gpibServer(GpibServer *);
    void processInt32(Int32Message *);
    void processChar8Array(Char8ArrayMessage *);
    bool getStartOk() const {return(startOk);}
private:
    const char *serverName;
    MessageServer *pMessageServer;
    Gpib* pGpib;
    bool startOk;
    int readBufSize;
    char *readBuf;
};

static char taskname[] = "gpibServer";
int initGpibServer(char *serverName, char *gpibName, 
		   int bufSize, int queueSize)
{
  if(bufSize<=0) {
    printf("initGpibServer %s bufSize must be >0\n",serverName);
    return(0);
  }
  GpibServer *psg =
    new GpibServer(serverName,gpibName,bufSize,queueSize);
  if(!psg->getStartOk()) return(0);
  int taskId = taskSpawn(taskname,100,VX_FP_TASK,2000,
			 (FUNCPTR)GpibServer::gpibServer,
			 (int)psg,0,0,0,0,0,0,0,0,0);
  if(taskId==ERROR) 
    printf("initGpibServer %s taskSpawn Failure\n",serverName);
  return(0);
}

GpibServer::GpibServer(const char *name, const char *gpibName, 
		       int bufSize, int queueSize)
: serverName(name), startOk(false), readBufSize(bufSize), readBuf(0)
{
    pGpib = Gpib::find(gpibName);
    if(pGpib==0) {
        printf("GpibServer: could not find Gpib %s\n",gpibName);
        return;
    }
    readBuf = (char *)(new char[(size_t)readBufSize]);
    pMessageServer = new MessageServer(name,queueSize);
    startOk = true;
}

void GpibServer::gpibServer(GpibServer *pgs)
{
  while(true) {
    pgs->pMessageServer->waitForMessage();
    Message *message;
    while((message = pgs->pMessageServer->receive())) {
      switch(message->getType()) {
      case messageTypeChar8Array:
	pgs->processChar8Array((Char8ArrayMessage *)message);
	break;
      case messageTypeInt32:
	pgs->processInt32((Int32Message *)message);
	break;
      default:
	printf("GpibServer %s recieved illegal message type %d\n",
	       pgs->serverName,message->getType());
	break;
      }
      delete message;
    }
  }
  return;
}

void GpibServer::processChar8Array(Char8ArrayMessage *preceive)
{
  int status=0;
  int actual=0;
  Char8ArrayMessage *preply = (Char8ArrayMessage *)pMessageServer->
    allocReplyMessage(preceive,messageTypeChar8Array);
  double timeout;
  if(preceive->timeoutUnits==timeoutUnitsSeconds) {
    timeout = preceive->timeout;
  } else if(preceive->timeoutUnits==timeoutUnitsMilliseconds) {
    timeout = (double)(preceive->timeout)/1000.0;
  } else {
     printf("GpibServer::processChar8Array: illegal timeoutUnits\n");
     timeout = preceive->timeout;
  }
  
  switch (preceive->cmd) {
  case gpibCmd:
    status = pGpib->cmd(preceive->value,
			preceive->getSize(),
			timeout);
    preply->status = status;
    //  printf ("GpibServer::processChar8Array gpibCmd status=%d\n",status);
    break;
  case gpibWrite:
    status = pGpib->write(preceive->value,
			  preceive->getSize(),
			  preceive->address,
			  timeout);
    preply->status = status;
    //  printf ("GpibServer::processChar8Array gpibWrite status=%d\n",status);
    break;
  case gpibRead:
    status = pGpib->read(readBuf,
			 readBufSize,&actual,
			 preceive->address,
			 timeout,
			 preceive->extra);
    preply->status = status;
    preply->allocValue(actual);
    ::memcpy((void *)preply->value,(void *)readBuf,actual);
    preply->setSize(actual);
    //  printf ("GpibServer::processChar8Array gpibRead status=%d\n",status);
    break;
  case gpibWriteRead:
    status = pGpib->write(preceive->value,
			  preceive->getSize(),
			  preceive->address,
			  timeout);
    //  printf ("GpibServer::processChar8Array gpibWriteRead write status=%d\n",status);
    if (status != preceive->getSize()) {
      preply->status = status;
    } else {
      status = pGpib->read(readBuf,
			   readBufSize,&actual,
			   preceive->address,
			   timeout,
			   preceive->extra);
      preply->status = status;
      preply->allocValue(actual);
      ::memcpy((void *)preply->value,(void *)readBuf,actual);
      preply->setSize(actual);
      //  printf ("GpibServer::processChar8Array gpibWriteRead read status=%d actual=%d\n",status, actual);
      
    }
    break;
  default:
    //  printf("GpibServer::processChar8Array %s illegal cmd %d\n", serverName, preceive->getType());
    break;
  }
  pMessageServer->reply(preply);
  return;
}

void GpibServer::processInt32(Int32Message *preceive)
{
  int status=0;
  int actual=0;
  Char8ArrayMessage *preply = (Char8ArrayMessage *)pMessageServer->
    allocReplyMessage(preceive,messageTypeChar8Array);
  double timeout;

  if(preceive->timeoutUnits==timeoutUnitsSeconds) {
    timeout = preceive->timeout;
  } else if(preceive->timeoutUnits==timeoutUnitsMilliseconds) {
    timeout = (double)(preceive->timeout)/1000.0;
  } else {
     printf("GpibServer::processChar8Array: illegal timeoutUnits\n");
     timeout = preceive->timeout;
  }
  switch (preceive->cmd) {
  case gpibRead:
    status = pGpib->read(readBuf,
			 readBufSize,&actual,
			 preceive->address,
			 timeout,
			 preceive->extra);
    preply->status = status;
    preply->allocValue(actual);
    ::memcpy((void *)preply->value,(void *)readBuf,actual);
    preply->setSize(actual);
    //  printf ("GpibServer::processInt32 gpibRead read status=%d actual=%d\n",status, actual);
    break;
  default:
    printf("GpibServer::processChar8Array %s illegal cmd %d\n", serverName, preceive->getType());
    break;
  }
  pMessageServer->reply(preply);
  return;
}

